# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import

import codecs
import os
import shutil
import zipfile

from anaconda_project.test.project_utils import project_no_dedicated_env
from anaconda_project.internal.test.tmpfile_utils import (with_directory_contents,
                                                          with_directory_contents_completing_project_file,
                                                          with_tmp_zipfile, complete_project_file_content)
from anaconda_project.test.environ_utils import minimal_environ
from anaconda_project.local_state_file import DEFAULT_LOCAL_STATE_FILENAME
from anaconda_project.local_state_file import LocalStateFile
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.requirements.download import DownloadRequirement
from anaconda_project.prepare import (prepare_without_interaction, unprepare, prepare_in_stages)
from anaconda_project import provide
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME

from tornado import gen

DATAFILE_CONTENT = ("downloads:\n"
                    "    DATAFILE:\n"
                    "        url: http://localhost/data.csv\n"
                    "        md5: 12345abcdef\n"
                    "        filename: data.csv\n")

ZIPPED_DATAFILE_CONTENT = ("downloads:\n"
                           "    DATAFILE:\n"
                           "        url: http://localhost/data.zip\n"
                           "        filename: data\n")

ZIPPED_DATAFILE_CONTENT_CHECKSUM = (ZIPPED_DATAFILE_CONTENT + "        md5: 12345abcdef\n")

ZIPPED_DATAFILE_CONTENT_NO_UNZIP = (ZIPPED_DATAFILE_CONTENT + "        unzip: false\n")

# have to specify unzip:true manually here
ZIPPED_DATAFILE_CONTENT_NO_ZIP_SUFFIX = ("downloads:\n"
                                         "    DATAFILE:\n"
                                         "        url: http://localhost/data\n"
                                         "        unzip: true\n"
                                         "        filename: data\n")


def _download_requirement():
    return DownloadRequirement(registry=RequirementsRegistry(),
                               env_var="DATAFILE",
                               url='http://localhost/data.csv',
                               filename='data.csv')


def test_prepare_and_unprepare_download(monkeypatch):
    def provide_download(dirname):
        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            with open(os.path.join(dirname, 'data.csv'), 'w') as out:
                out.write('data')
            self._hash = '12345abcdef'
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)
        project = project_no_dedicated_env(dirname)
        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert hasattr(result, 'environ')
        assert 'DATAFILE' in result.environ
        filename = os.path.join(dirname, 'data.csv')
        assert os.path.exists(filename)

        project.frontend.reset()
        status = unprepare(project, result)
        assert project.frontend.logs == ["Removed downloaded file %s." % filename,
                                         ("Current environment is not in %s, no need to delete it." % dirname)]
        assert status.status_description == 'Success.'
        assert status
        assert not os.path.exists(filename)

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: DATAFILE_CONTENT}, provide_download)


def test_prepare_download_mismatched_checksum_after_download(monkeypatch):
    def provide_download(dirname):
        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            with open(os.path.join(dirname, 'data.csv'), 'w') as out:
                out.write('data')
            self._hash = 'mismatched'
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)

        project = project_no_dedicated_env(dirname)
        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert not result
        assert ('Error downloading http://localhost/data.csv: mismatched hashes. '
                'Expected: 12345abcdef, calculated: mismatched') in result.errors

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: DATAFILE_CONTENT}, provide_download)


def test_prepare_download_exception(monkeypatch):
    def provide_download(dirname):
        @gen.coroutine
        def mock_downloader_run(self, loop):
            raise Exception('error')

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)
        project = project_no_dedicated_env(dirname)
        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert not result
        assert ('missing requirement to run this project: A downloaded file which is referenced by DATAFILE.'
                ) in result.errors

        project.frontend.reset()
        status = unprepare(project, result)
        filename = os.path.join(dirname, 'data.csv')
        assert project.frontend.logs == ["No need to remove %s which wasn't downloaded." % filename,
                                         ("Current environment is not in %s, no need to delete it." % dirname)]
        assert status.status_description == 'Success.'

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: DATAFILE_CONTENT}, provide_download)


def test_unprepare_download_fails(monkeypatch):
    def provide_download(dirname):
        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            with open(os.path.join(dirname, 'data.csv'), 'w') as out:
                out.write('data')
            self._hash = '12345abcdef'
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)
        project = project_no_dedicated_env(dirname)
        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert hasattr(result, 'environ')
        assert 'DATAFILE' in result.environ
        filename = os.path.join(dirname, 'data.csv')
        assert os.path.exists(filename)

        def mock_remove(path):
            raise IOError("Not gonna remove this")

        monkeypatch.setattr("os.remove", mock_remove)

        project.frontend.reset()
        status = unprepare(project, result)
        assert project.frontend.logs == []
        assert status.status_description == ('Failed to remove %s: Not gonna remove this.' % filename)
        assert status.errors == []
        assert not status
        assert os.path.exists(filename)

        monkeypatch.undo()  # so os.remove isn't broken during directory cleanup

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: DATAFILE_CONTENT}, provide_download)


def test_provide_minimal(monkeypatch):
    MIN_DATAFILE_CONTENT = ("downloads:\n" "    DATAFILE: http://localhost/data.csv\n")

    def provide_download(dirname):
        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            with open(os.path.join(dirname, 'data.csv'), 'w') as out:
                out.write('data')
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)
        project = project_no_dedicated_env(dirname)
        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert hasattr(result, 'environ')
        assert 'DATAFILE' in result.environ

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: MIN_DATAFILE_CONTENT}, provide_download)


def test_provide_no_download_in_check_mode(monkeypatch):
    MIN_DATAFILE_CONTENT = ("downloads:\n" "    DATAFILE: http://localhost/data.csv\n")

    def provide_download(dirname):
        @gen.coroutine
        def mock_downloader_run(self, loop):
            raise Exception("should not have tried to download in check mode")

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)

        project = project_no_dedicated_env(dirname)
        result = prepare_without_interaction(project,
                                             environ=minimal_environ(PROJECT_DIR=dirname),
                                             mode=provide.PROVIDE_MODE_CHECK)
        assert not result

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: MIN_DATAFILE_CONTENT}, provide_download)


def test_provide_missing_url(monkeypatch):
    ERR_DATAFILE_CONTENT = ("downloads:\n" "    DATAFILE:\n" "       filename: data.csv\n")

    def provide_download(dirname):
        project = project_no_dedicated_env(dirname)
        prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert "Download item DATAFILE doesn't contain a 'url' field." in project.problems

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ERR_DATAFILE_CONTENT}, provide_download)


def test_provide_empty_url(monkeypatch):
    ERR_DATAFILE_CONTENT = ("downloads:\n" "    DATAFILE:\n" "       url: \"\"\n")

    def provide_download(dirname):
        project = project_no_dedicated_env(dirname)
        prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert "Download item DATAFILE has an empty 'url' field." in project.problems

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ERR_DATAFILE_CONTENT}, provide_download)


def test_provide_multiple_checksums(monkeypatch):
    ERR_DATAFILE_CONTENT = ("downloads:\n"
                            "    DATAFILE:\n"
                            "       url: http://localhost/\n"
                            "       md5: abcdefg\n"
                            "       sha1: abcdefg\n")

    def provide_download(dirname):
        project = project_no_dedicated_env(dirname)
        prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert "Multiple checksums for download DATAFILE: md5 and sha1." in project.problems

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ERR_DATAFILE_CONTENT}, provide_download)


def test_provide_wrong_form(monkeypatch):
    ERR_DATAFILE_CONTENT = ("downloads:\n" "    - http://localhost/data.csv\n")

    def provide_download(dirname):
        project = project_no_dedicated_env(dirname)
        prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert ("%s: 'downloads:' section should be a dictionary, found ['http://localhost/data.csv']" %
                DEFAULT_PROJECT_FILENAME) in project.problems

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ERR_DATAFILE_CONTENT}, provide_download)


def test_failed_download(monkeypatch):
    def provide_download(dirname):
        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 400
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)
        project = project_no_dedicated_env(dirname)
        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert not result
        assert ('missing requirement to run this project: A downloaded file which is referenced by DATAFILE.'
                ) in result.errors

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: DATAFILE_CONTENT}, provide_download)


def test_failed_download_before_connect(monkeypatch):
    def provide_download(dirname):
        @gen.coroutine
        def mock_downloader_run(self, loop):
            # if we don't even get an HTTP response, the errors are handled this way,
            # e.g. if the URL is bad.
            self._errors = ['This went horribly wrong']
            raise gen.Return(None)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)
        project = project_no_dedicated_env(dirname)
        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert not result
        assert ('missing requirement to run this project: A downloaded file which is referenced by DATAFILE.'
                ) in result.errors

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: DATAFILE_CONTENT}, provide_download)


def test_file_exists(monkeypatch):
    def provide_download(dirname):
        FILENAME = os.path.join(dirname, 'data.csv')
        requirement = _download_requirement()
        local_state_file = LocalStateFile.load_for_directory(dirname)
        local_state_file.set_service_run_state(requirement.env_var, {'filename': FILENAME})
        local_state_file.save()
        with open(FILENAME, 'w') as out:
            out.write('data')
        project = project_no_dedicated_env(dirname)

        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert hasattr(result, 'environ')
        assert 'DATAFILE' in result.environ

    LOCAL_STATE = ("DATAFILE:\n" "  filename: data.csv")

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: DATAFILE_CONTENT,
            DEFAULT_LOCAL_STATE_FILENAME: LOCAL_STATE
        }, provide_download)


def test_prepare_download_of_zip_file(monkeypatch):
    def provide_download_of_zip(zipname, dirname):
        with codecs.open(os.path.join(dirname, DEFAULT_PROJECT_FILENAME), 'w', 'utf-8') as f:
            f.write(complete_project_file_content(ZIPPED_DATAFILE_CONTENT))

        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            assert self._url.endswith(".zip")
            assert self._filename.endswith(".zip")
            shutil.copyfile(zipname, self._filename)
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)

        project = project_no_dedicated_env(dirname)

        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert hasattr(result, 'environ')
        assert 'DATAFILE' in result.environ
        assert os.path.isdir(os.path.join(dirname, 'data'))
        assert os.path.isfile(os.path.join(dirname, 'data', 'foo'))
        assert codecs.open(os.path.join(dirname, 'data', 'foo')).read() == 'hello\n'

    with_tmp_zipfile(dict(foo='hello\n'), provide_download_of_zip)


def test_prepare_download_of_zip_file_checksum(monkeypatch):
    def provide_download_of_zip(zipname, dirname):
        with codecs.open(os.path.join(dirname, DEFAULT_PROJECT_FILENAME), 'w', 'utf-8') as f:
            f.write(complete_project_file_content(ZIPPED_DATAFILE_CONTENT_CHECKSUM))

        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            assert self._url.endswith(".zip")
            assert self._filename.endswith(".zip")
            shutil.copyfile(zipname, self._filename)
            self._hash = '12345abcdef'
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)
        project = project_no_dedicated_env(dirname)

        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert hasattr(result, 'environ')
        assert 'DATAFILE' in result.environ
        assert os.path.isdir(os.path.join(dirname, 'data'))
        assert os.path.isfile(os.path.join(dirname, 'data', 'foo'))
        assert codecs.open(os.path.join(dirname, 'data', 'foo')).read() == 'hello\n'

        project.frontend.reset()
        status = unprepare(project, result)
        filename = os.path.join(dirname, 'data')
        assert project.frontend.logs == ["Removed downloaded file %s." % filename,
                                         ("Current environment is not in %s, no need to delete it." % dirname)]
        assert status.status_description == "Success."

    with_tmp_zipfile(dict(foo='hello\n'), provide_download_of_zip)


def test_prepare_download_of_zip_file_no_unzip(monkeypatch):
    def provide_download_of_zip_no_unzip(zipname, dirname):
        with codecs.open(os.path.join(dirname, DEFAULT_PROJECT_FILENAME), 'w', 'utf-8') as f:
            f.write(complete_project_file_content(ZIPPED_DATAFILE_CONTENT_NO_UNZIP))

        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            assert self._url.endswith(".zip")
            # we aren't going to unzip so we should be downloading straignt to
            # the specified filename 'data' without the .zip on it
            assert not self._filename.endswith(".zip")
            shutil.copyfile(zipname, self._filename)
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)

        project = project_no_dedicated_env(dirname)

        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert hasattr(result, 'environ')
        assert 'DATAFILE' in result.environ
        assert os.path.isfile(os.path.join(dirname, 'data'))
        with zipfile.ZipFile(os.path.join(dirname, 'data')) as zf:
            assert zf.namelist() == ['foo']

    with_tmp_zipfile(dict(foo='hello\n'), provide_download_of_zip_no_unzip)


def test_prepare_download_of_zip_file_no_zip_extension(monkeypatch):
    def provide_download_of_zip(zipname, dirname):
        with codecs.open(os.path.join(dirname, DEFAULT_PROJECT_FILENAME), 'w', 'utf-8') as f:
            f.write(complete_project_file_content(ZIPPED_DATAFILE_CONTENT_NO_ZIP_SUFFIX))

        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            # we add .zip to the download filename, even though it wasn't in the URL
            assert not self._url.endswith(".zip")
            assert self._filename.endswith(".zip")
            shutil.copyfile(zipname, self._filename)
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)

        project = project_no_dedicated_env(dirname)

        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert hasattr(result, 'environ')
        assert 'DATAFILE' in result.environ
        assert os.path.isdir(os.path.join(dirname, 'data'))
        assert os.path.isfile(os.path.join(dirname, 'data', 'foo'))
        assert codecs.open(os.path.join(dirname, 'data', 'foo')).read() == 'hello\n'

    with_tmp_zipfile(dict(foo='hello\n'), provide_download_of_zip)


def test_prepare_download_of_broken_zip_file(monkeypatch):
    def provide_download_of_zip(dirname):
        with codecs.open(os.path.join(dirname, DEFAULT_PROJECT_FILENAME), 'w', 'utf-8') as f:
            f.write(complete_project_file_content(ZIPPED_DATAFILE_CONTENT))

        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            assert self._url.endswith(".zip")
            assert self._filename.endswith(".zip")
            with codecs.open(self._filename, 'w', 'utf-8') as f:
                f.write("This is not a zip file.")
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)

        project = project_no_dedicated_env(dirname)

        result = prepare_without_interaction(project, environ=minimal_environ(PROJECT_DIR=dirname))
        assert not result
        assert [("Failed to unzip %s: File is not a zip file" % os.path.join(dirname, "data.zip")),
                "missing requirement to run this project: A downloaded file which is referenced by DATAFILE.",
                "  Environment variable DATAFILE is not set."] == result.errors

    with_directory_contents(dict(), provide_download_of_zip)


def _download_status(prepare_context):
    for status in prepare_context.statuses:
        if isinstance(status.requirement, DownloadRequirement):
            return status
    return None


def test_configure(monkeypatch):
    def check(dirname):
        @gen.coroutine
        def mock_downloader_run(self, loop):
            class Res:
                pass

            res = Res()
            res.code = 200
            with open(os.path.join(dirname, 'data.csv'), 'w') as out:
                out.write('data')
            self._hash = '12345abcdef'
            raise gen.Return(res)

        monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)

        project = project_no_dedicated_env(dirname)
        environ = minimal_environ(PROJECT_DIR=dirname)
        stage = prepare_in_stages(project, environ=environ)
        status = None
        while status is None and stage is not None:
            prepare_context = stage.configure()
            status = _download_status(prepare_context)
            if status is None:
                stage = stage.execute()

        assert status is not None

        req = status.requirement
        provider = status.provider

        # check initial config

        config = provider.read_config(req, prepare_context.environ, prepare_context.local_state_file,
                                      prepare_context.default_env_spec_name, prepare_context.overrides)

        assert dict(source='download') == config

        config['source'] = 'environ'
        config['value'] = 'abc.txt'

        provider.set_config_values_as_strings(req, prepare_context.environ, prepare_context.local_state_file,
                                              prepare_context.default_env_spec_name, prepare_context.overrides, config)

        config = provider.read_config(req, prepare_context.environ, prepare_context.local_state_file,
                                      prepare_context.default_env_spec_name, prepare_context.overrides)

        assert dict(source='download', value='abc.txt') == config

        config['source'] = 'variables'
        config['value'] = 'qrs.txt'
        provider.set_config_values_as_strings(req, prepare_context.environ, prepare_context.local_state_file,
                                              prepare_context.default_env_spec_name, prepare_context.overrides, config)

        config = provider.read_config(req, prepare_context.environ, prepare_context.local_state_file,
                                      prepare_context.default_env_spec_name, prepare_context.overrides)

        assert dict(source='variables', value='qrs.txt') == config

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
downloads:
  FOO: http://example.com/data.csv
    """}, check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os

from anaconda_project.internal import conda_api

# we keep the conda env variables because otherwise
# we'd have to keep recreating project-specific conda envs in our tests
system_vars_to_keep = ('PATH',
                       'LD_LIBRARY_PATH',
                       'TERM',
                       'PYTHONPATH',
                       'HOME',
                       # Windows stuff
                       'SystemRoot',
                       'SystemDrive',
                       'OS',
                       'ProgramData',
                       'ProgramFiles',
                       'LOCALAPPDATA',
                       'HOMEDRIVE',
                       'HOMEPATH')
conda_vars_to_keep = ('CONDA_DEFAULT_ENV', 'CONDA_ENV_PATH', 'CONDA_PREFIX')


def _minimal_environ_full(with_conda_env, **additions):
    minimal_environ = dict()
    for name in (system_vars_to_keep + conda_vars_to_keep):
        if name in os.environ:
            minimal_environ[name] = os.environ[name]

    env_prefix = conda_api.environ_get_prefix(minimal_environ)

    if len(additions) > 0 or not with_conda_env:
        if not with_conda_env:
            for name in conda_vars_to_keep:
                if name in minimal_environ:
                    del minimal_environ[name]

        for (key, value) in additions.items():
            minimal_environ[key] = value

    return (env_prefix, minimal_environ)


def minimal_environ(**additions):
    """Get an environment with minimal likely weird side effects on tests, while still working."""
    return _minimal_environ_full(with_conda_env=True, **additions)[1]


def minimal_environ_no_conda_env(**additions):
    """Get a minimal environment without the conda env in it."""
    return _minimal_environ_full(with_conda_env=False, **additions)[1]


def env_prefix_and_minimal_environ(**additions):
    """Get a tuple of env_prefix and minimal environ without conda vars."""
    return _minimal_environ_full(with_conda_env=False, **additions)


def strip_environ(environ):
    """Pull system variables back out of our minimal environ so we can check test results without noise."""
    copy = environ.copy()
    for name in (system_vars_to_keep + conda_vars_to_keep):
        if name in copy:
            del copy[name]
    return copy


def strip_environ_keeping_conda_env(environ):
    """Pull system variables back out of our minimal environ so we can check test results without noise."""
    copy = environ.copy()
    for name in system_vars_to_keep:
        if name in copy:
            del copy[name]
    return copy



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
import anaconda_project.requirements_registry.network_util as network_util

import socket


def test_can_connect_to_socket():
    # create a listening socket just to get a port number
    # that (probably) won't be in use after we close it
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("127.0.0.1", 0))
    s.listen(1)
    port = s.getsockname()[1]

    try:
        assert network_util.can_connect_to_socket("127.0.0.1", port)
    finally:
        s.close()


def test_cannot_connect_to_socket():
    # create a listening socket just to get a port number
    # that (probably) won't be in use after we close it
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("127.0.0.1", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()

    assert not network_util.can_connect_to_socket("127.0.0.1", port)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import pytest

from anaconda_project.internal.cli.command_commands import main
from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents_completing_project_file
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.project import Project


class Args(object):
    def __init__(self, type, name, command, directory='.'):
        self.type = type
        self.name = name
        self.command = command
        self.directory = directory
        self.env_spec = None
        self.supports_http_options = None


def test_add_command_ask_type(monkeypatch):
    def check_ask_type(dirname):
        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        def mock_console_input(prompt):
            return "b"

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        args = Args(None, 'test', 'file.py', directory=dirname)
        res = main(args)
        assert res == 0

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert len(command.keys()) == 2
        assert command['bokeh_app'] == 'file.py'
        assert command['env_spec'] == 'default'

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check_ask_type)


def test_add_command_not_interactive(monkeypatch, capsys):
    def check(dirname):
        def mock_is_interactive():
            return False

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        args = Args(None, 'test', 'file.py', directory=dirname)
        res = main(args)
        assert res == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert 'Specify the --type option to add this command.\n' == err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check)


def test_add_command_ask_type_interrupted(monkeypatch, capsys):
    def check_ask_type(dirname):
        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        def mock_input(prompt):
            raise KeyboardInterrupt('^C')

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils._input', mock_input)

        args = Args(None, 'test', 'file.py', directory=dirname)
        with pytest.raises(SystemExit) as excinfo:
            main(args)
        assert excinfo.value.code == 1

        out, err = capsys.readouterr()
        assert out == ''
        assert err == '\nCanceling\n\n'

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check_ask_type)


def test_add_command_ask_other_shell(monkeypatch):
    def check(dirname):
        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        def mock_console_input(prompt):
            return "c"

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        def mock_system():
            return "Linux"

        monkeypatch.setattr('platform.system', mock_system)

        args = Args(None, 'test', 'echo hello', directory=dirname)
        res = main(args)
        assert res == 0

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert len(command.keys()) == 2
        assert command['unix'] == 'echo hello'
        assert command['env_spec'] == 'default'

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check)


def test_add_command_ask_other_windows(monkeypatch):
    def check(dirname):
        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        def mock_console_input(prompt):
            return "c"

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        def mock_system():
            return "Windows"

        monkeypatch.setattr('platform.system', mock_system)

        args = Args(None, 'test', 'echo hello', directory=dirname)
        res = main(args)
        assert res == 0

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert len(command.keys()) == 2
        assert command['windows'] == 'echo hello'
        assert command['env_spec'] == 'default'

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check)


def test_add_command_ask_type_twice(monkeypatch, capsys):
    def check_ask_type(dirname):
        def mock_is_interactive():
            return True

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

        calls = []

        def mock_console_input(prompt):
            res = ['-', 'b'][len(calls)]
            calls.append(True)
            return res

        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.console_input', mock_console_input)

        args = Args(None, 'test', 'file.py', directory=dirname)
        res = main(args)
        assert res == 0
        assert len(calls) == 2

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert len(command.keys()) == 2
        assert command['bokeh_app'] == 'file.py'
        assert command['env_spec'] == 'default'
        out, err = capsys.readouterr()
        assert out == ("Please enter 'b', 'n', or 'c'.\n" +
                       "    A Bokeh app is the project-relative path to a Bokeh script or app directory.\n" +
                       "    A notebook file is the project-relative path to a .ipynb file.\n"
                       "    A command line is any command you might type at the command prompt.\n"
                       "Added a command 'test' to the project. Run it with `anaconda-project run test`.\n")

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check_ask_type)


def test_add_command_specifying_notebook(monkeypatch, capsys):
    def check_specifying_notebook(dirname):
        args = Args('notebook', 'test', 'file.ipynb', directory=dirname)
        res = main(args)
        assert res == 0

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert command['notebook'] == 'file.ipynb'
        assert command['env_spec'] == 'default'
        assert len(command.keys()) == 2

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: 'packages:\n - notebook\n'}, check_specifying_notebook)


def test_add_command_guessing_notebook(monkeypatch, capsys):
    def check_guessing_notebook(dirname):
        args = Args(None, 'test', 'file.ipynb', directory=dirname)
        res = main(args)
        assert res == 0

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert command['notebook'] == 'file.ipynb'
        assert command['env_spec'] == 'default'
        assert len(command.keys()) == 2

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: 'packages:\n - notebook\n',
         'file.ipynb': "{}"}, check_guessing_notebook)


def test_add_command_with_env_spec(monkeypatch, capsys):
    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'add-command', '--directory', dirname, '--env-spec',
                                               'foo', '--type', 'notebook', 'test', 'file.ipynb'])
        assert code == 0

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert command['notebook'] == 'file.ipynb'
        assert command['env_spec'] == 'foo'
        assert len(command.keys()) == 2

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: 'env_specs:\n  default: {}\n  foo: {}\n'}, check)


def test_add_command_with_supports_http_options(monkeypatch, capsys):
    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'add-command', '--directory', dirname,
                                               '--supports-http-options', '--type', 'notebook', 'test', 'file.ipynb'])
        assert code == 0

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert command['notebook'] == 'file.ipynb'
        assert command['env_spec'] == 'default'
        assert command['supports_http_options'] is True
        assert len(command.keys()) == 3

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: 'env_specs:\n  default: {}\n  foo: {}\n'}, check)


def test_add_command_with_no_supports_http_options(monkeypatch, capsys):
    def check(dirname):
        code = _parse_args_and_run_subcommand(
            ['anaconda-project', 'add-command', '--directory', dirname, '--no-supports-http-options', '--type',
             'notebook', 'test', 'file.ipynb'])
        assert code == 0

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert command['notebook'] == 'file.ipynb'
        assert command['env_spec'] == 'default'
        assert command['supports_http_options'] is False
        assert len(command.keys()) == 3

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: 'env_specs:\n  default: {}\n  foo: {}\n'}, check)


def _test_command_command_project_problem(capsys, monkeypatch, command, append_dir=False):
    def check(dirname):
        if append_dir:
            command.extend(['--directory', dirname])
        code = _parse_args_and_run_subcommand(command)
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert ('variables section contains wrong value type 42,' + ' should be dict or list of requirements\n' +
                'Unable to load the project.\n') in err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: "variables:\n  42"}, check)


def test_add_command_project_problem(capsys, monkeypatch):
    _test_command_command_project_problem(capsys,
                                          monkeypatch,
                                          [
                                              'anaconda-project', 'add-command', '--type', 'notebook', 'test',
                                              'file.ipynb'
                                          ],
                                          append_dir=True)


def test_add_command_breaks_project(capsys, monkeypatch):
    def check_problem_add_cmd(dirname):
        args = Args('notebook', 'test', 'file.ipynb', directory=dirname)
        res = main(args)
        assert res == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert (("%s: command 'test' has multiple commands in it, 'notebook' can't go with 'unix'\n" %
                 DEFAULT_PROJECT_FILENAME) + "Unable to add the command.\n") == err

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: ("commands:\n  test:\n    unix: foo\n")}, check_problem_add_cmd)


def test_remove_command_with_project_file_problems(capsys, monkeypatch):
    _test_command_command_project_problem(capsys,
                                          monkeypatch,
                                          ['anaconda-project', 'remove-command', 'test'],
                                          append_dir=True)


def test_remove_command(monkeypatch, capsys):
    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-command', 'test', '--directory', dirname])
        assert code == 0

        project = Project(dirname)

        command = project.project_file.get_value(['commands', 'test'])
        assert command is None

        out, err = capsys.readouterr()
        assert out == "Removed the command 'test' from the project.\n"
        assert err == ''

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: 'packages: ["notebook"]\ncommands:\n  test:\n    notebook: file.ipynb\n'}, check)


def test_remove_command_missing(monkeypatch, capsys):
    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-command', 'test', '--directory', dirname])
        assert code == 1

        out, err = capsys.readouterr()
        assert err == "Command: 'test' not found in project file.\n"
        assert out == ''

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check)


def test_list_commands_with_project_file_problems(capsys, monkeypatch):
    _test_command_command_project_problem(capsys, monkeypatch, ['anaconda-project', 'list-commands'], append_dir=True)


def test_list_commands_empty_project(capsys):
    def check_empty_project(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-commands', '--directory', dirname])
        assert code == 0

        out, err = capsys.readouterr()
        assert '' == err
        assert ("No commands found for project: {}\n\n".format(dirname)) == out

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ""}, check_empty_project)


def test_list_commands(capsys):
    def check_empty_project(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-commands', '--directory', dirname])
        assert code == 0

        out, err = capsys.readouterr()
        assert '' == err

        expected_out = """
Commands for project: {dirname}

Name          Description
====          ===========
default       Bokeh app test.py
run_notebook  Notebook test.ipynb
""".format(dirname=dirname).strip() + "\n"

        assert expected_out == out

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: ("commands:\n"
                                    "  default:\n"
                                    "    bokeh_app: test.py\n"
                                    "  run_notebook:\n"
                                    "    notebook: test.ipynb\n"
                                    "packages:\n"
                                    " - bokeh\n"
                                    " - notebook\n")}, check_empty_project)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import

from copy import deepcopy
import os
import platform
import pytest
import subprocess
import sys

from anaconda_project.test.environ_utils import minimal_environ, strip_environ
from anaconda_project.test.project_utils import project_no_dedicated_env
from anaconda_project.internal.test.tmpfile_utils import (with_directory_contents,
                                                          with_directory_contents_completing_project_file)
from anaconda_project.internal import conda_api
from anaconda_project.prepare import (prepare_without_interaction, unprepare, prepare_in_stages, PrepareSuccess,
                                      PrepareFailure, _after_stage_success, _FunctionPrepareStage)
from anaconda_project.project import Project
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.project_commands import ProjectCommand
from anaconda_project.requirements_registry.requirement import UserConfigOverrides
from anaconda_project.conda_manager import (push_conda_manager_class, pop_conda_manager_class, CondaManager,
                                            CondaEnvironmentDeviations, CondaLockSet)


@pytest.mark.slow
def test_prepare_empty_directory():
    def prepare_empty(dirname):
        project = Project(dirname)
        environ = minimal_environ()
        result = prepare_without_interaction(project, environ=environ)
        assert result.errors == []
        assert result
        assert result.env_prefix is not None
        assert dict(PROJECT_DIR=project.directory_path) == strip_environ(result.environ)
        assert dict() == strip_environ(environ)
        assert result.command_exec_info is None

    with_directory_contents(dict(), prepare_empty)


def test_prepare_bad_provide_mode():
    def prepare_bad_provide_mode(dirname):
        with pytest.raises(ValueError) as excinfo:
            project = project_no_dedicated_env(dirname)
            environ = minimal_environ()
            prepare_in_stages(project, mode="BAD_PROVIDE_MODE", environ=environ)
        assert "invalid provide mode" in repr(excinfo.value)

    with_directory_contents(dict(), prepare_bad_provide_mode)


@pytest.mark.slow
@pytest.mark.skipif(platform.system() == 'Windows' and
                    not (sys.version_info.major == 3 and sys.version_info.minor == 4),
                    reason="on Windows, can't delete env dir except on python 3.4, don't know why")
def test_unprepare_empty_directory():
    def unprepare_empty(dirname):
        project = Project(dirname)
        environ = minimal_environ()
        result = prepare_without_interaction(project, environ=environ)
        assert result.errors == []
        assert result
        status = unprepare(project, result)
        assert status.errors == []
        assert status

    with_directory_contents(dict(), unprepare_empty)


def test_unprepare_problem_project():
    def unprepare_problems(dirname):
        project = project_no_dedicated_env(dirname)
        environ = minimal_environ()
        result = prepare_without_interaction(project, environ=environ)
        assert not result
        assert result.env_prefix is None
        status = unprepare(project, result)
        assert not status
        assert status.status_description == 'Unable to load the project.'
        assert status.errors == [('%s: variables section contains wrong value type 42, ' +
                                  'should be dict or list of requirements') % project.project_file.basename]

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: "variables:\n  42"}, unprepare_problems)


@pytest.mark.slow
def test_unprepare_nothing_to_do():
    def unprepare_nothing(dirname):
        project = Project(dirname)
        environ = minimal_environ()
        result = prepare_without_interaction(project, environ=environ)
        assert result.errors == []
        assert result
        status = unprepare(project, result, whitelist=[])
        assert status.errors == []
        assert status
        assert status.status_description == 'Nothing to clean up.'

    with_directory_contents(dict(), unprepare_nothing)


def test_default_to_system_environ():
    def prepare_system_environ(dirname):
        project = project_no_dedicated_env(dirname)
        os_environ_copy = deepcopy(os.environ)
        result = prepare_without_interaction(project)
        assert project.directory_path == strip_environ(result.environ)['PROJECT_DIR']
        # os.environ wasn't modified
        assert os_environ_copy == os.environ
        # result.environ inherits everything in os.environ
        for key in os_environ_copy:
            if key == 'PATH' and platform.system() == 'Windows' and result.environ[key] != os.environ[key]:
                print("prepare changed PATH on Windows and ideally it would not.")
            else:
                if key == 'PATH' and result.environ[key] != os.environ[key]:
                    original = os.environ[key].split(os.pathsep)
                    updated = result.environ[key].split(os.pathsep)
                    print("ORIGINAL PATH: " + repr(original))
                    print("UPDATED PATH: " + repr(updated))
                    assert original == updated
                assert result.errors == []
                assert result
                assert result.environ.get(key) == os.environ.get(key)

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
packages: []
        """
        }, prepare_system_environ)


def test_prepare_some_env_var_already_set():
    def prepare_some_env_var(dirname):
        project = project_no_dedicated_env(dirname)
        environ = minimal_environ(FOO='bar')
        result = prepare_without_interaction(project, environ=environ)
        assert result.errors == []
        assert result
        assert dict(FOO='bar', PROJECT_DIR=project.directory_path) == strip_environ(result.environ)
        assert dict(FOO='bar') == strip_environ(environ)

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: {}
"""}, prepare_some_env_var)


def test_prepare_some_env_var_not_set():
    def prepare_some_env_var(dirname):
        project = project_no_dedicated_env(dirname)
        environ = minimal_environ(BAR='bar')
        result = prepare_without_interaction(project, environ=environ)
        assert not result
        assert result.env_prefix is not None
        assert dict(BAR='bar') == strip_environ(environ)

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: {}
"""}, prepare_some_env_var)


def test_prepare_some_env_var_not_set_keep_going():
    def prepare_some_env_var_keep_going(dirname):
        project = project_no_dedicated_env(dirname)
        environ = minimal_environ(BAR='bar')
        stage = prepare_in_stages(project, environ=environ, keep_going_until_success=True)
        assert "Set up project." == stage.description_of_action
        assert ['FOO', 'CONDA_PREFIX'] == [status.requirement.env_var for status in stage.statuses_before_execute]

        # there's an initial stage to set the conda env
        next_stage = stage.execute()
        assert ['FOO', 'CONDA_PREFIX'] == [status.requirement.env_var for status in stage.statuses_after_execute]
        assert not stage.failed
        assert stage.environ['PROJECT_DIR'] == dirname
        assert "Set up project." == next_stage.description_of_action
        assert ['FOO', 'CONDA_PREFIX'] == [status.requirement.env_var for status in next_stage.statuses_before_execute]
        stage = next_stage

        for i in range(1, 10):
            next_stage = stage.execute()
            assert next_stage is not None
            assert stage.failed
            assert stage.environ['PROJECT_DIR'] == dirname
            stage = next_stage
        assert dict(BAR='bar') == strip_environ(environ)

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: {}
"""}, prepare_some_env_var_keep_going)


def test_prepare_with_app_entry():
    def prepare_with_app_entry(dirname):
        project = project_no_dedicated_env(dirname)
        environ = minimal_environ(FOO='bar')
        env_path = conda_api.environ_get_prefix(environ)
        result = prepare_without_interaction(project, environ=environ)
        assert result

        command = result.command_exec_info
        assert 'FOO' in command.env
        assert command.cwd == project.directory_path
        if platform.system() == 'Windows':
            commandpath = os.path.join(env_path, "python.exe")
        else:
            commandpath = os.path.join(env_path, "bin", "python")
        assert command.args == [commandpath, 'echo.py', env_path, 'foo', 'bar']
        p = command.popen(stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        # strip is to pull off the platform-specific newline
        assert out.decode().strip() == ("['echo.py', '%s', 'foo', 'bar']" % (env_path.replace("\\", "\\\\")))
        assert err.decode() == ""

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: {}

commands:
  default:
    conda_app_entry: python echo.py ${PREFIX} foo bar
""",
         "echo.py": """
from __future__ import print_function
import sys
print(repr(sys.argv))
"""}, prepare_with_app_entry)


def test_prepare_choose_command():
    def check(dirname):
        project = project_no_dedicated_env(dirname)
        environ = minimal_environ()
        result = prepare_without_interaction(project, environ=environ, command_name='foo')
        assert result.errors == []
        assert result
        assert os.path.join(project.directory_path, 'foo.py') in result.command_exec_info.args

        environ = minimal_environ()
        result = prepare_without_interaction(project, environ=environ, command_name='bar')
        assert result.errors == []
        assert result
        assert os.path.join(project.directory_path, 'bar.py') in result.command_exec_info.args

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
commands:
    foo:
       bokeh_app: foo.py
    bar:
       bokeh_app: bar.py
packages:
  - bokeh
""",
         "foo.py": "# foo",
         "bar.py": "# bar"}, check)


def test_prepare_command_not_in_project():
    def check(dirname):
        # create a command that isn't in the Project
        project = project_no_dedicated_env(dirname)
        command = ProjectCommand(name="foo",
                                 attributes=dict(bokeh_app="foo.py",
                                                 env_spec=project.default_env_spec_name))
        environ = minimal_environ()
        result = prepare_without_interaction(project, environ=environ, command=command)
        assert result.errors == []
        assert result
        assert os.path.join(project.directory_path, 'foo.py') in result.command_exec_info.args

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
commands:
  decoy:
    description: "do not use me"
    unix: foobar
    windows: foobar
""",
         "foo.py": "# foo"}, check)


def test_prepare_bad_command_name():
    def check(dirname):
        project = project_no_dedicated_env(dirname)
        environ = minimal_environ(BAR='bar')
        result = prepare_without_interaction(project, environ=environ, command_name="blah")
        assert not result
        assert result.env_prefix is None
        assert result.errors
        assert "Command name 'blah' is not in" in result.errors[0]

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: """
"""}, check)


def _push_fake_env_creator():
    class HappyCondaManager(CondaManager):
        def __init__(self, frontend):
            pass

        def resolve_dependencies(self, package_specs, channels, platforms):
            return CondaLockSet({})

        def find_environment_deviations(self, prefix, spec):
            return CondaEnvironmentDeviations(summary="all good",
                                              missing_packages=(),
                                              wrong_version_packages=(),
                                              missing_pip_packages=(),
                                              wrong_version_pip_packages=())

        def fix_environment_deviations(self, prefix, spec, deviations=None, create=True):
            pass

        def remove_packages(self, prefix, packages):
            pass

    push_conda_manager_class(HappyCondaManager)


def _pop_fake_env_creator():
    pop_conda_manager_class()


def test_prepare_choose_environment():
    def check(dirname):
        env_var = conda_api.conda_prefix_variable()

        try:
            _push_fake_env_creator()
            project = Project(dirname)
            environ = minimal_environ()
            result = prepare_without_interaction(project, environ=environ, env_spec_name='foo')
            expected_path = project.env_specs['foo'].path(project.directory_path)
            assert result.environ[env_var] == expected_path

            environ = minimal_environ()
            result = prepare_without_interaction(project, environ=environ, env_spec_name='bar')
            assert result.errors == []
            assert result
            expected_path = project.env_specs['bar'].path(project.directory_path)
            assert result.environ[env_var] == expected_path
        finally:
            _pop_fake_env_creator()

    with_directory_contents(
        {DEFAULT_PROJECT_FILENAME: """
name: blah
platforms: [linux-32,linux-64,osx-64,win-32,win-64]
env_specs:
    foo: {}
    bar: {}
"""}, check)


def test_prepare_use_command_specified_env_spec():
    def check(dirname):
        env_var = conda_api.conda_prefix_variable()

        try:
            _push_fake_env_creator()
            project = Project(dirname)
            environ = minimal_environ()
            # we specify the command name but not the
            # env_spec_name but it should imply the proper env
            # spec name.
            result = prepare_without_interaction(project, environ=environ, command_name='hello')
            expected_path = project.env_specs['foo'].path(project.directory_path)
            assert result.environ[env_var] == expected_path
        finally:
            _pop_fake_env_creator()

    with_directory_contents(
        {DEFAULT_PROJECT_FILENAME: """
name: blah
platforms: [linux-32,linux-64,osx-64,win-32,win-64]
env_specs:
    default: {}
    foo: {}
    bar: {}
commands:
    hello:
       env_spec: foo
       unix: echo hello
       windows: echo hello
"""}, check)


def test_update_environ():
    def prepare_then_update_environ(dirname):
        project = project_no_dedicated_env(dirname)
        environ = minimal_environ(FOO='bar')
        result = prepare_without_interaction(project, environ=environ)
        assert result.errors == []
        assert result

        other = minimal_environ(BAR='baz')
        result.update_environ(other)
        assert dict(FOO='bar', BAR='baz', PROJECT_DIR=dirname) == strip_environ(other)

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: {}
"""}, prepare_then_update_environ)


def test_attempt_to_grab_result_early():
    def early_result_grab(dirname):
        project = project_no_dedicated_env(dirname)
        first_stage = prepare_in_stages(project)
        with pytest.raises(RuntimeError) as excinfo:
            first_stage.result
        assert "result property isn't available" in repr(excinfo.value)

    with_directory_contents(dict(), early_result_grab)


def test_attempt_to_grab_statuses_early():
    def early_status_grab(dirname):
        project = project_no_dedicated_env(dirname)
        first_stage = prepare_in_stages(project)
        with pytest.raises(RuntimeError) as excinfo:
            first_stage.statuses_after_execute
        assert "statuses_after_execute isn't available" in repr(excinfo.value)

    with_directory_contents(dict(), early_status_grab)


def test_skip_after_success_function_when_second_stage_fails():
    state = {'state': 'start'}

    def do_first(stage):
        assert state['state'] == 'start'
        state['state'] = 'first'
        stage.set_result(
            PrepareSuccess(statuses=(),
                           command_exec_info=None,
                           environ=dict(),
                           overrides=UserConfigOverrides(),
                           env_spec_name='first'),
            [])

        def last(stage):
            assert state['state'] == 'first'
            state['state'] = 'second'
            stage.set_result(
                PrepareFailure(statuses=(),
                               errors=[],
                               environ=dict(),
                               overrides=UserConfigOverrides(),
                               env_spec_name='last'),
                [])
            return None

        return _FunctionPrepareStage(dict(), UserConfigOverrides(), "second", [], last)

    first_stage = _FunctionPrepareStage(dict(), UserConfigOverrides(), "first", [], do_first)

    def after(updated_statuses):
        raise RuntimeError("should not have been called")

    stage = _after_stage_success(first_stage, after)
    assert stage.overrides is first_stage.overrides
    assert isinstance(stage.environ, dict)
    while stage is not None:
        next_stage = stage.execute()
        result = stage.result
        if result.failed:
            assert stage.failed
            break
        else:
            assert not stage.failed
        stage = next_stage
    assert result.failed
    assert state['state'] == 'second'


def test_run_after_success_function_when_second_stage_succeeds():
    state = {'state': 'start'}

    def do_first(stage):
        assert state['state'] == 'start'
        state['state'] = 'first'
        stage.set_result(
            PrepareSuccess(statuses=(),
                           command_exec_info=None,
                           environ=dict(),
                           overrides=UserConfigOverrides(),
                           env_spec_name='foo'),
            [])

        def last(stage):
            assert state['state'] == 'first'
            state['state'] = 'second'
            stage.set_result(
                PrepareSuccess(statuses=(),
                               command_exec_info=None,
                               environ=dict(),
                               overrides=UserConfigOverrides(),
                               env_spec_name='bar'),
                [])
            return None

        return _FunctionPrepareStage(dict(), UserConfigOverrides(), "second", [], last)

    first_stage = _FunctionPrepareStage(dict(), UserConfigOverrides(), "first", [], do_first)

    def after(updated_statuses):
        assert state['state'] == 'second'
        state['state'] = 'after'

    stage = _after_stage_success(first_stage, after)

    assert stage.overrides is first_stage.overrides
    assert stage.description_of_action == first_stage.description_of_action
    assert stage.environ == first_stage.environ
    assert stage.statuses_before_execute is first_stage.statuses_before_execute
    stage.configure()  # checking it doesn't raise

    while stage is not None:
        next_stage = stage.execute()

        if hasattr(stage, '_stage'):
            assert stage.statuses_after_execute is stage._stage.statuses_after_execute
            assert stage.failed is stage._stage.failed

        result = stage.result
        if result.failed:
            assert stage.failed
            break
        else:
            assert not stage.failed
        stage = next_stage
    assert not result.failed
    assert state['state'] == 'after'


def _monkeypatch_download_file(monkeypatch, dirname, filename='MYDATA', checksum=None):
    from tornado import gen

    @gen.coroutine
    def mock_downloader_run(self, loop):
        class Res:
            pass

        res = Res()
        res.code = 200
        with open(os.path.join(dirname, filename), 'w') as out:
            out.write('data')
        if checksum:
            self._hash = checksum
        raise gen.Return(res)

    monkeypatch.setattr("anaconda_project.internal.http_client.FileDownloader.run", mock_downloader_run)


def test_provide_whitelist(monkeypatch):
    def check(dirname):
        from anaconda_project.requirements_registry.requirements.conda_env import CondaEnvRequirement

        _monkeypatch_download_file(monkeypatch, dirname, filename="nope")

        no_foo = [('missing requirement to run this project: A downloaded file which is ' + 'referenced by FOO.'),
                  '  Environment variable FOO is not set.']

        # whitelist only the env req by class
        project = project_no_dedicated_env(dirname)
        environ = minimal_environ()
        result = prepare_without_interaction(project, provide_whitelist=(CondaEnvRequirement, ), environ=environ)
        assert result.errors == no_foo

        # whitelist by instance
        env_req = [req for req in project.requirements(None) if isinstance(req, CondaEnvRequirement)][0]
        result = prepare_without_interaction(project, provide_whitelist=(env_req, ), environ=environ)
        assert result.errors == no_foo

        # whitelist by variable name
        result = prepare_without_interaction(project, provide_whitelist=(env_req.env_var, ), environ=environ)
        assert result.errors == no_foo

        # whitelist the download
        result = prepare_without_interaction(project,
                                             provide_whitelist=(env_req, project.download_requirements(None)[0]),
                                             environ=environ)
        assert result.errors == []
        assert 'FOO' in result.environ

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
downloads:
  FOO: "http://example.com/nope"

"""}, check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function, unicode_literals

import codecs
import os
import pytest

from anaconda_project.internal.cli.bug_handler import handle_bugs


def test_no_bug_happens(capsys):
    def unbuggy_main():
        print("hi")
        return 0

    code = handle_bugs(unbuggy_main, program_name="myprogram", details_dict=dict())
    assert code is 0

    out, err = capsys.readouterr()

    assert 'hi\n' == out
    assert '' == err


def test_bug_handling(capsys, monkeypatch):
    def mock_is_interactive():
        return True

    monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

    def buggy_main():
        raise AssertionError("It did not work")

    # we name the program something wonky to be sure we slugify
    # it, and use a non-BMP unicode char to be sure we handle that
    code = handle_bugs(buggy_main,
                       program_name=u"my$ 🌟program",
                       details_dict=dict(thing1="foo",
                                         thing2="bar",
                                         thing3=u"🌟"))

    assert code is 1
    out, err = capsys.readouterr()

    assert '' == out
    assert err.startswith(u"""An unexpected error occurred, most likely a bug in my$ 🌟program.
    (The error was: AssertionError: It did not work)
Details about the error were saved to """)
    filename = err.split()[-1]
    assert filename.endswith(".txt")
    assert os.path.basename(filename).startswith("bug_details_my---program_")
    assert os.path.isfile(filename)

    with codecs.open(filename, 'r', 'utf-8') as f:
        lines = f.readlines()
        report = "".join(lines)

    # we can't easily test the exact value (it contains line
    # numbers of full file paths), but check the major items are
    # in there.
    assert report.startswith(u"Bug details for my$ 🌟program error on ")
    assert 'sys.argv' in report
    assert 'test_bug_handler.py' in report
    assert 'Traceback' in report
    assert 'AssertionError' in report
    assert 'It did not work' in report

    os.remove(filename)


def test_bug_handling_not_interactive(capsys, monkeypatch):
    def mock_is_interactive():
        return False

    monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

    def buggy_main():
        raise AssertionError("It did not work")

    # we name the program something wonky to be sure we slugify
    # it, and use a non-BMP unicode char to be sure we handle that
    code = handle_bugs(buggy_main,
                       program_name=u"my$ 🌟program",
                       details_dict=dict(thing1="foo",
                                         thing2="bar",
                                         thing3=u"🌟"))

    assert code is 1
    out, err = capsys.readouterr()

    assert '' == out

    assert u"""An unexpected error occurred, most likely a bug in my$ 🌟program.
    (The error was: AssertionError: It did not work)""" in err
    assert u"Bug details for my$ 🌟program" in err
    assert u"sys.argv:" in err
    assert u"Traceback" in err
    assert u"raise AssertionError" in err
    assert u"Above details were also saved to" in err

    filename = err.split()[-1]
    assert filename.endswith(".txt")
    assert os.path.basename(filename).startswith("bug_details_my---program_")
    assert os.path.isfile(filename)

    with codecs.open(filename, 'r', 'utf-8') as f:
        lines = f.readlines()
        report = "".join(lines)

    # we can't easily test the exact value (it contains line
    # numbers of full file paths), but check the major items are
    # in there.
    assert report.startswith(u"Bug details for my$ 🌟program error on ")
    assert 'sys.argv' in report
    assert 'test_bug_handler.py' in report
    assert 'Traceback' in report
    assert 'AssertionError' in report
    assert 'It did not work' in report

    os.remove(filename)


def test_bug_handling_is_buggy(capsys, monkeypatch):
    def buggy_main():
        raise AssertionError("It did not work")

    def mock_temporary_file(*args, **kwargs):
        raise IOError("no temporary file")

    monkeypatch.setattr('tempfile.NamedTemporaryFile', mock_temporary_file)

    with pytest.raises(AssertionError) as exc_info:
        handle_bugs(buggy_main, program_name=u"myprogram", details_dict=dict())
    assert 'It did not work' == str(exc_info.value)

    out, err = capsys.readouterr()

    assert '' == out
    assert """An unexpected error occurred, most likely a bug in myprogram.
    (The error was: AssertionError: It did not work)
""" == err


def test_keyboard_interrupt(capsys):
    def buggy_main():
        raise KeyboardInterrupt("ctrl-c")

    code = handle_bugs(buggy_main, program_name=u"myprogram", details_dict=dict())

    assert code is 1
    out, err = capsys.readouterr()

    assert '' == out
    assert 'myprogram was interrupted.\n' == err



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os
import zipfile

from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.internal.test.tmpfile_utils import (with_directory_contents,
                                                          with_directory_contents_completing_project_file)
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME


def test_archive_command_on_empty_project(capsys):
    def check(dirname):
        archivefile = os.path.join(dirname, "foo.zip")
        code = _parse_args_and_run_subcommand(['anaconda-project', 'archive', '--directory', dirname, archivefile])
        assert code == 1

        out, err = capsys.readouterr()
        assert "anaconda-project.yml does not exist.\nCan't create an archive.\n" == err
        assert '' == out

        assert not os.path.exists(os.path.join(dirname, DEFAULT_PROJECT_FILENAME))
        assert not os.path.exists(archivefile)

    with_directory_contents(dict(), check)


def test_archive_command_on_simple_project(capsys):
    def check(dirname):
        archivefile = os.path.join(dirname, "foo.zip")
        code = _parse_args_and_run_subcommand(['anaconda-project', 'archive', '--directory', dirname, archivefile])
        assert code == 0

        unlocked_warning = ("Warning: env specs are not locked, which means they may not work "
                            "consistently for others or when deployed.\n"
                            "  Consider using the 'anaconda-project lock' command to lock the project.")

        out, err = capsys.readouterr()
        assert ('  added %s\n  added %s\n%s\nCreated project archive %s\n' %
                (os.path.join("some_name", DEFAULT_PROJECT_FILENAME), os.path.join("some_name", "foo.py"),
                 unlocked_warning, archivefile)) == out

        with zipfile.ZipFile(archivefile, mode='r') as zf:
            assert [os.path.basename(x) for x in sorted(zf.namelist())] == [DEFAULT_PROJECT_FILENAME, "foo.py"]

        assert os.path.exists(os.path.join(dirname, DEFAULT_PROJECT_FILENAME))

        assert '' == err

    with_directory_contents_completing_project_file({'foo.py': 'print("hello")\n'}, check)


def test_archive_command_on_invalid_project(capsys):
    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'archive', '--directory', dirname, 'foo.zip'])
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert ('variables section contains wrong value type 42,' + ' should be dict or list of requirements\n' +
                'Unable to load the project.\n') in err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: "variables:\n  42"}, check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Types related to project requirement providers."""
from __future__ import absolute_import

from abc import ABCMeta, abstractmethod
from copy import deepcopy
import os
import shutil

from anaconda_project.internal import conda_api
from anaconda_project.internal import logged_subprocess
from anaconda_project.internal.metaclass import with_metaclass
from anaconda_project.internal.makedirs import makedirs_ok_if_exists
from anaconda_project.internal.simple_status import SimpleStatus


def _service_directory(local_state_file, relative_name):
    return os.path.join(os.path.dirname(local_state_file.filename), "services", relative_name)


class ProvideContext(object):
    """A context passed to ``Provider.provide()`` representing state that can be modified."""

    def __init__(self, environ, local_state_file, default_env_spec_name, status, mode, frontend):
        """Create a ProvideContext.

        Args:
            environ (dict): environment variables to be read and modified
            local_state_file (LocalStateFile): to store any created state
            status (RequirementStatus): current status
            mode (str): one of PROVIDE_MODE_PRODUCTION, PROVIDE_MODE_DEVELOPMENT, PROVIDE_MODE_CHECK
        """
        self.environ = environ
        self._local_state_file = local_state_file
        self._default_env_spec_name = default_env_spec_name
        self._status = status
        self._mode = mode
        self._frontend = frontend

    def ensure_service_directory(self, relative_name):
        """Create a directory in PROJECT_DIR/services with the given name.

        The name should be unique to the ServiceRequirement creating the directory,
        so usually the requirement's env var.

        Args:
            relative_name (str): name to distinguish this dir from other service directories
        """
        path = _service_directory(self._local_state_file, relative_name)
        makedirs_ok_if_exists(path)
        return path

    def transform_service_run_state(self, service_name, func):
        """Run a function which takes and potentially modifies the state of a service.

        If the function modifies the state it's given, the new state will be saved
        and passed in next time.

        Args:
            service_name (str): the name of the service, should be
                specific enough to uniquely identify the provider
            func (function): function to run, passing it the current state

        Returns:
            Whatever ``func`` returns.
        """
        old_state = self._local_state_file.get_service_run_state(service_name)
        modified = deepcopy(old_state)
        result = func(modified)
        if modified != old_state:
            self._local_state_file.set_service_run_state(service_name, modified)
            self._local_state_file.save()
        return result

    @property
    def status(self):
        """Get the current ``RequirementStatus``."""
        return self._status

    @property
    def local_state_file(self):
        """Get the LocalStateFile."""
        return self._local_state_file

    @property
    def default_env_spec_name(self):
        """Get the default env spec."""
        return self._default_env_spec_name

    @property
    def mode(self):
        """Get flavor of provide.

        Value should be ``PROVIDE_MODE_DEVELOPMENT``, ``PROVIDE_MODE_PRODUCTION``, or ``PROVIDE_MODE_CHECK``.
        """
        return self._mode

    @property
    def frontend(self):
        """Get the current ``Frontend``."""
        return self._frontend


def shutdown_service_run_state(local_state_file, service_name):
    """Run any shutdown commands from the local state file for the given service.

    Also remove the shutdown commands from the file.

    Args:
        local_state_file (LocalStateFile): local state
        service_name (str): the name of the service, usually a
            variable name, should be specific enough to uniquely
            identify the provider

    Returns:
        a `Status` instance potentially containing errors
    """
    run_states = local_state_file.get_all_service_run_states()
    if service_name not in run_states:
        return SimpleStatus(success=True, description=("Nothing to do to shut down %s." % service_name))

    errors = []
    state = run_states[service_name]
    if 'shutdown_commands' in state:
        commands = state['shutdown_commands']
        for command in commands:
            code = logged_subprocess.call(command)
            if code != 0:
                errors.append("Shutting down %s, command %s failed with code %d." % (service_name, repr(command), code))
    # clear out the run state once we try to shut it down
    local_state_file.set_service_run_state(service_name, dict())
    local_state_file.save()

    if errors:
        return SimpleStatus(success=False,
                            description=("Shutdown commands failed for %s." % service_name),
                            errors=errors)
    else:
        return SimpleStatus(success=True, description=("Successfully shut down %s." % service_name))


def delete_service_directory(local_state_file, relative_name):
    """Delete a directory in PROJECT_DIR/services with the given name.

    The name should be unique to the ServiceRequirement creating the directory,
    so usually the requirement's env var.

    IF this fails, it does so silently (returns no errors).

    Args:
        relative_name (str): name to distinguish this dir from other service directories

    Returns:
        None
    """
    path = _service_directory(local_state_file, relative_name)
    try:
        shutil.rmtree(path=path)
    except OSError:
        pass
    # also delete the services directory itself, if it's now empty
    try:
        # this fails on non-empty dir
        os.rmdir(os.path.dirname(path))
    except OSError:
        pass


class ProviderAnalysis(object):
    """A Provider's preflight check snapshotting the state prior to ``provide()``.

    Instances of this class are immutable, and are usually created as part of a
    ``RequirementStatus``.
    """

    def __init__(self, config, missing_env_vars_to_configure, missing_env_vars_to_provide):
        """Create a ProviderAnalysis."""
        self._config = deepcopy(config)  # defensive copy so we don't modify the original
        self._missing_env_vars_to_configure = missing_env_vars_to_configure
        self._missing_env_vars_to_provide = missing_env_vars_to_provide

    @property
    def config(self):
        """Get the configuration dict from the time of analysis."""
        return self._config

    @property
    def missing_env_vars_to_configure(self):
        """Get the env vars we were missing in order to configure, from the time of analysis."""
        return self._missing_env_vars_to_configure

    @property
    def missing_env_vars_to_provide(self):
        """Get the env vars we were missing in order to provide, from the time of analysis."""
        return self._missing_env_vars_to_provide


class ProvideResult(object):
    """A Provider's results from the ``provide()`` call.

    Instances of this class are immutable, and are returned from ``provide()``.
    """

    def __init__(self, errors=None):
        """Create a ProvideResult."""
        if errors is None:
            errors = []
        self._errors = errors

    def copy_with_additions(self, errors=None):
        """Copy this result, appending additional errors."""
        if errors is None:
            errors = []
        if len(errors) == 0:
            # we don't have to actually copy since we are immutable
            return self
        else:
            return ProvideResult(errors=(self._errors + errors))

    @property
    def errors(self):
        """Get any fatal errors that occurred during provide() preventing success."""
        return self._errors

    @classmethod
    def empty(cls):
        """Get an empty ProvideResult (currently a singleton since these are immutable)."""
        return _empty_provide_result

# get this via ProvideResult.empty()
_empty_provide_result = ProvideResult()


class Provider(with_metaclass(ABCMeta)):
    """A Provider can take some action to meet a Requirement."""

    @abstractmethod
    def missing_env_vars_to_configure(self, requirement, environ, local_state_file):
        """Get a list of unset environment variable names that must be set before configuring this provider.

        Args:
            requirement (Requirement): requirement instance we are providing for
            environ (dict): current environment variable dict
            local_state_file (LocalStateFile): local state file
        """
        pass  # pragma: no cover

    @abstractmethod
    def missing_env_vars_to_provide(self, requirement, environ, local_state_file):
        """Get a list of unset environment variable names that must be set before calling provide().

        Args:
            requirement (Requirement): requirement instance we are providing for
            environ (dict): current environment variable dict
            local_state_file (LocalStateFile): local state file
        """
        pass  # pragma: no cover

    @abstractmethod
    def read_config(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
        """Read a config dict from the local state file for the given requirement.

        You can think of this as the GET returning a web form for
        configuring the provider. And in fact it was once used for
        that, though we deleted the html stuff now.

        The returned 'config' has a 'source' field which was
        essentially a selected radio option for where to get the
        requirement, and other fields are entry boxes underneath
        each radio option.

        This method still exists in the code in case we want to
        do a textual version (or a new HTML version, but probably
        outside of the anaconda-project codebase). See also
        UI_MODE_TEXT_ASK_QUESTIONS in the cli code.

        Args:
            requirement (Requirement): the requirement we're providing
            environ (dict): current environment variables
            local_state_file (LocalStateFile): file to read from
            default_env_spec_name (str): the fallback env spec name
            overrides (UserConfigOverrides): user-supplied forced config

        """
        pass  # pragma: no cover

    def set_config_values_as_strings(self, requirement, environ, local_state_file, default_env_spec_name, overrides,
                                     values):
        """Set some config values in the state file (should not save the file).

        You can think of this as the POST submitting a web form
        for configuring the provider. And in fact it was once used
        for that, though we deleted the html stuff now.

        Args:
            requirement (Requirement): the requirement we're providing
            environ (dict): current environment variables
            local_state_file (LocalStateFile): file to save to
            default_env_spec_name (str): default env spec name for this prepare
            overrides (UserConfigOverrides): if any values in here change, delete the override
            values (dict): dict from string to string

        """
        pass  # silently ignore unknown config values

    def analyze(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
        """Analyze whether and how we'll be able to provide the requirement.

        This is used to show the situation in the UI, and also to
        consolidate all IO-type work in one place (inside
        Requirement.check_status()).

        Returns:
          A ``ProviderAnalysis`` instance.
        """
        config = self.read_config(requirement, environ, local_state_file, default_env_spec_name, overrides)
        missing_to_configure = self.missing_env_vars_to_configure(requirement, environ, local_state_file)
        missing_to_provide = self.missing_env_vars_to_provide(requirement, environ, local_state_file)
        return ProviderAnalysis(config=config,
                                missing_env_vars_to_configure=missing_to_configure,
                                missing_env_vars_to_provide=missing_to_provide)

    @abstractmethod
    def provide(self, requirement, context):
        """Execute the provider, fulfilling the requirement.

        The implementation should read and modify the passed-in
        ``environ`` rather than accessing the OS environment
        directly.

        Args:
            requirement (Requirement): requirement we want to meet
            context (ProvideContext): context containing project state

        Returns:
            a ``ProvideResult`` instance

        """
        pass  # pragma: no cover

    @abstractmethod
    def unprovide(self, requirement, environ, local_state_file, overrides, requirement_status=None):
        """Undo the provide, cleaning up any files or processes we created.

        The requirement may still be met after this, if our providing wasn't
        really needed.

        Args:
            requirement (Requirement): requirement we want to de-provide
            environ (dict): current env vars, often from a previous prepare
            local_state_file (LocalStateFile): the local state
            overrides (UserConfigOverrides): overrides to state
            requirement_status (RequirementStatus or None): requirement status if available

        Returns:
            a `Status` instance describing the (non)success of the unprovision
        """
        pass  # pragma: no cover


class EnvVarProvider(Provider):
    """Meets a requirement for an env var by letting people set it manually."""

    def _local_state_override(self, requirement, local_state_file):
        return local_state_file.get_value(["variables", requirement.env_var], default=None)

    def _disabled_local_state_override(self, requirement, local_state_file):
        return local_state_file.get_value(["disabled_variables", requirement.env_var], default=None)

    def missing_env_vars_to_configure(self, requirement, environ, local_state_file):
        """Override superclass to require env prefix."""
        if self._get_env_prefix(environ) is not None:
            return ()
        else:
            return (conda_api.conda_prefix_variable(), )

    def missing_env_vars_to_provide(self, requirement, environ, local_state_file):
        """Override superclass to require env prefix."""
        return self.missing_env_vars_to_configure(requirement, environ, local_state_file)

    def _get_env_prefix(self, environ):
        # on unix, ENV_PATH is the prefix and DEFAULT_ENV can be just a name,
        # on windows DEFAULT_ENV is always the prefix
        return environ.get(conda_api.conda_prefix_variable(), None)

    def read_config(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
        """Override superclass to read env var value."""
        config = dict()
        value = None
        if requirement.encrypted:
            # import keyring locally because it's an optional dependency
            # that prints a warning when it's needed but not found.
            import anaconda_project.internal.keyring as keyring

            env_prefix = self._get_env_prefix(environ)
            if env_prefix is None:
                value = None
            else:
                value = keyring.get(env_prefix, requirement.env_var)

        # note that we will READ an encrypted value from local
        # state if someone puts it in there by hand, but we won't
        # ever write one there ourselves.
        if value is None:
            value = self._local_state_override(requirement, local_state_file)

        disabled_value = self._disabled_local_state_override(requirement, local_state_file)
        was_disabled = value is None and disabled_value is not None
        if was_disabled:
            value = disabled_value

        if value is not None:
            config['value'] = value

        if value is not None and not was_disabled:
            source = 'variables'
        elif requirement.env_var in environ:
            source = 'environ'
            config['value'] = environ[requirement.env_var]
        elif 'default' in requirement.options:
            source = 'default'
        else:
            source = 'unset'
        config['source'] = source
        return config

    def set_config_values_as_strings(self, requirement, environ, local_state_file, default_env_spec_name, overrides,
                                     values):
        """Override superclass to set env var value."""
        if requirement.encrypted:
            self._set_encrypted_config_values_as_strings(requirement, environ, local_state_file, default_env_spec_name,
                                                         overrides, values)
        else:
            self._set_nonencrypted_config_values_as_strings(requirement, environ, local_state_file,
                                                            default_env_spec_name, overrides, values)

    def _set_nonencrypted_config_values_as_strings(self, requirement, environ, local_state_file, default_env_spec_name,
                                                   overrides, values):
        override_path = ["variables", requirement.env_var]
        disabled_path = ["disabled_variables", requirement.env_var]

        # we set override_path only if the source is variables,
        # otherwise the value goes in disabled_variables. If we
        # don't have a source that means the only option we
        # presented was to set the local override, so default to
        # 'variables'
        overriding = (values.get('source', 'variables') == 'variables')

        # If there's an existing override value and the source is not 'variables',
        # we need to be sure to move the existing to disabled_variables.
        # Also, we save values['value'] from the web form in local_state_file,
        # even if we aren't using it as the source right now.

        local_override_value = self._local_state_override(requirement, local_state_file)
        if local_override_value is None:
            local_override_value = self._disabled_local_state_override(requirement, local_state_file)

        value_string = values.get('value', local_override_value)

        if value_string is not None:
            if value_string == '':
                # the reason empty string unsets is that otherwise there's no easy
                # way to unset from a web form
                local_state_file.unset_value(override_path)
                local_state_file.unset_value(disabled_path)
            else:
                if overriding:
                    local_state_file.set_value(override_path, value_string)
                    local_state_file.unset_value(disabled_path)
                else:
                    local_state_file.set_value(disabled_path, value_string)
                    local_state_file.unset_value(override_path)

    def _set_encrypted_config_values_as_strings(self, requirement, environ, local_state_file, default_env_spec_name,
                                                overrides, values):
        # import keyring locally because it's an optional dependency
        # that prints a warning when it's needed but not found.
        import anaconda_project.internal.keyring as keyring

        env_prefix = self._get_env_prefix(environ)
        from_keyring = keyring.get(env_prefix, requirement.env_var)
        value_string = values.get('value', from_keyring)

        if value_string is not None:
            if value_string == '':
                keyring.unset(env_prefix, requirement.env_var)
            else:
                keyring.set(env_prefix, requirement.env_var, value_string)

    def provide(self, requirement, context):
        """Override superclass to use configured env var (or already-set env var)."""
        # We prefer the values in this order:
        #  - value set in project-local state overrides everything
        #    (otherwise the UI for configuring the value would end
        #    up ignored)
        #  - value in the keyring overrides (treated the same as
        #    anaconda-project-local.yml, but for encrypted variables)
        #  - then anything already set in the environment wins, so you
        #    can override on the command line like `FOO=bar myapp`
        #  - then the anaconda-project.yml default value
        local_state_override = None
        if requirement.encrypted:
            # import keyring locally because it's an optional dependency
            # that prints a warning when it's needed but not found.
            import anaconda_project.internal.keyring as keyring

            env_prefix = self._get_env_prefix(context.environ)
            if env_prefix is not None:
                local_state_override = keyring.get(env_prefix, requirement.env_var)

        # we will read encrypted vars from local state, though we never
        # put them in there ourselves
        if local_state_override is None:
            local_state_override = self._local_state_override(requirement, context.local_state_file)

        if local_state_override is not None:
            # anaconda-project-local.yml
            #
            # variables:
            #   REDIS_URL: "redis://example.com:1234"
            context.environ[requirement.env_var] = local_state_override
        elif requirement.env_var in context.environ:
            # nothing to do here
            pass
        elif 'default' in requirement.options:
            # anaconda-project.yml
            #
            # variables:
            #   REDIS_URL:
            #     default: "redis://example.com:1234"
            value = requirement.options['default']
            if value is not None:
                context.environ[requirement.env_var] = value
        else:
            pass

        return ProvideResult.empty().copy_with_additions(errors=[])

    def unprovide(self, requirement, environ, local_state_file, overrides, requirement_status=None):
        """Override superclass to return success always."""
        return SimpleStatus(success=True, description=("Nothing to clean up for %s." % requirement.env_var))



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import collections
import subprocess
import os
import re
import sys

from anaconda_project.internal import logged_subprocess


class PipError(Exception):
    """General pip error."""

    pass


class PipNotInstalledError(PipError):
    """Pip isn't even installed in this environment."""

    pass


# this function exists so we can monkeypatch it in tests
def _get_pip_command(prefix, extra_args):
    # we need to use the pip from the prefix
    unix_pip = os.path.join(prefix, "bin", "pip")
    win_pip = os.path.join(prefix, "Scripts", "pip.exe")
    pips = [pip for pip in [unix_pip, win_pip] if os.path.exists(pip)]
    if len(pips) == 0:
        raise PipNotInstalledError("'pip' command is not installed in the environment %s" % prefix)
    cmd_list = [pips[0]]
    cmd_list.extend(extra_args)
    return cmd_list


def _call_pip(prefix, extra_args):
    cmd_list = _get_pip_command(prefix, extra_args)

    try:
        p = logged_subprocess.Popen(cmd_list, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as e:
        raise PipError("failed to run: %r: %r" % (" ".join(cmd_list), repr(e)))
    (out, err) = p.communicate()
    errstr = err.decode().strip()
    if p.returncode != 0:
        raise PipError('%s: %s' % (" ".join(cmd_list), errstr))
    elif errstr != '':
        for line in errstr.split("\n"):
            print("%s %s: %s" % (cmd_list[0], cmd_list[1], line), file=sys.stderr)
    return out


def install(prefix, pkgs=None):
    """Install packages into an environment."""
    if not pkgs or not isinstance(pkgs, (list, tuple)):
        raise TypeError('must specify a list of one or more packages to install into existing environment, not %r' %
                        pkgs)

    # --no-deps is because we don't want to pull in pip versions of
    # everything that conda has.
    args = ['install', '--quiet', '--no-deps']
    args.extend(pkgs)

    return _call_pip(prefix, extra_args=args)


def remove(prefix, pkgs=None):
    """Remove packages from an environment."""
    if not pkgs or not isinstance(pkgs, (list, tuple)):
        raise TypeError('must specify a list of one or more packages to remove from existing environment')

    args = ['uninstall', '--quiet', '--yes']
    args.extend(pkgs)
    return _call_pip(prefix, extra_args=args)


def installed(prefix):
    """Get a dict of package names to (name, version) tuples."""
    if not os.path.isdir(prefix):
        return dict()

    # In pip 9, there's a big ugly deprecation warning by default if
    # you type `pip list`, unless you do `pip list --format=legacy`
    # pip 8 of course does not support --format=legacy, so that
    # can't be done unconditionally.
    format_args = ["--format=legacy"]
    try:
        out = _call_pip(prefix, extra_args=['--version']).decode('utf-8')

        if out.startswith("pip "):
            try:
                major_version = int(out[4])
                if major_version <= 8:
                    format_args = []
            except ValueError:
                # didn't get an integer, who knows
                pass
    except PipNotInstalledError:
        pass

    try:
        out = _call_pip(prefix, extra_args=(['list'] + format_args)).decode('utf-8')
        # on Windows, $ in a regex doesn't match \r\n, we need to get rid of \r
        out = out.replace("\r\n", "\n")
    except PipNotInstalledError:
        out = ""  # if pip isn't installed, there are no pip packages
    # the output to parse ("legacy" format mode) looks like this:
    #   ympy (0.7.6.1)
    #   tables (3.2.2)
    #   terminado (0.5)
    line_re = re.compile("^ *([^ ]+) *\(([^)]+)\)$", flags=re.MULTILINE)
    result = dict()
    for match in line_re.finditer(out):
        result[match.group(1)] = (match.group(1), match.group(2))
    return result


ParsedPipSpec = collections.namedtuple('ParsedPipSpec', ['name'])

_spec_pat = re.compile(' *([a-zA-Z0-9][-_.a-zA-Z0-9]+)')

_egg_fragment_re = re.compile('[#&]egg=([^&]*)')

_egg_fragment_postfix_re = re.compile('^(.*?)(?:-dev|-\d.*)$')

_url_schemes = set(('http', 'https', 'file', 'ftp', 'git', 'git+http', 'git+https', 'git+ssh', 'git+git', 'git+file',
                    'hg', 'hg+http', 'hg+https', 'hg+ssh', 'hg+static-http'
                    'bzr', 'bzr+http', 'bzr+https', 'bzr+ssh', 'bzr+sftp', 'bzr+ftp', 'bzr+lp', 'svn', 'svn+ssh',
                    'svn+http', 'svn+https', 'svn+svn'))


def _is_pip_understood_url(s):
    if ':' in s:
        scheme = s.split(':', 1)[0]
        return scheme.lower() in _url_schemes
    else:
        return False


def _extract_name(spec):
    m = _spec_pat.match(spec)
    if m is not None:
        return m.group(1)
    else:
        return None


def _extract_name_from_egg_fragment(url):
    m = _egg_fragment_re.search(url)
    if m is not None:
        fragment = _extract_name(m.group(1))
        if fragment is not None:
            # it's allowed to put "#egg=foo-1.2" but then pip just
            # ignores the "-1.2"
            m = _egg_fragment_postfix_re.search(fragment)
            if m is not None:
                return m.group(1)
            else:
                return fragment
    else:
        return None


def parse_spec(spec):
    """Parse a pip spec, right now we only understand the name portion.

    Parsing it exactly as pip would is extremely complicated, and
    we would pretty much have to import the pip code.
    So for now we'll "fail late" when we invoke pip.

    What we understand currently is an url with a "#egg=foo"
    fragment, and a plain package name (possibly with version info
    on it). We don't understand filesystem paths yet.

    Returns:
       ``ParsedPipSpec`` or None on failure

    """
    if _is_pip_understood_url(spec):
        name = _extract_name_from_egg_fragment(spec)
    else:
        name = _extract_name(spec)

    if name is None:
        return None
    else:
        return ParsedPipSpec(name=name)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import collections
import platform
import sys

_PY2 = sys.version_info[0] == 2


def is_unicode(s):
    if _PY2:  # pragma: no cover (py2/py3)
        return isinstance(s, unicode)  # pragma: no cover (py2/py3) # noqa
    else:  # pragma: no cover (py2/py3)
        return isinstance(s, str)  # pragma: no cover (py2/py3)


def is_string(s):
    if _PY2:  # pragma: no cover (py2/py3)
        return isinstance(s, basestring)  # pragma: no cover (py2/py3) # noqa
    else:  # pragma: no cover (py2/py3)
        return isinstance(s, str)  # pragma: no cover (py2/py3)


def is_list(v):
    return isinstance(v, collections.Sequence) and not is_string(v)


def is_dict(v):
    return isinstance(v, collections.Mapping)


def env_without_unicode(environ):
    # On Windows / Python 2.7, Popen explodes if given unicode strings in the environment.
    if _PY2 and platform.system() == 'Windows':  # pragma: no cover (py2/py3)
        environ_copy = dict()
        for key, value in environ.items():
            if isinstance(key, unicode):  # noqa
                key = key.encode()
            if isinstance(value, unicode):  # noqa
                value = value.encode()
            assert isinstance(key, str)
            assert isinstance(value, str)
            environ_copy[key] = value
        return environ_copy
    else:  # pragma: no cover (py2/py3)
        return environ



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Command-line-specific project load utilities."""
from __future__ import absolute_import, print_function

import sys

from anaconda_project.project import Project
from anaconda_project.frontend import Frontend

import anaconda_project.internal.cli.console_utils as console_utils


class CliFrontend(Frontend):
    def __init__(self):
        super(CliFrontend, self).__init__()

    def info(self, message):
        print(message)

    def error(self, message):
        print(message, file=sys.stderr)

    def partial_info(self, data):
        sys.stdout.write(data)
        sys.stdout.flush()

    def partial_error(self, data):
        sys.stderr.write(data)
        sys.stderr.flush()


def load_project(dirname):
    """Load a Project, fixing it if needed and possible."""
    project = Project(dirname, frontend=CliFrontend())

    if console_utils.stdin_is_interactive():
        had_fixable = len(project.fixable_problems) > 0
        for problem in project.fixable_problems:
            print(problem.text)
            should_fix = console_utils.console_ask_yes_or_no(problem.fix_prompt, default=False)
            if should_fix:
                problem.fix(project)
            else:
                problem.no_fix(project)

        # both fix() and no_fix() can modify project_file, if no changes
        # were made this is a no-op.
        if had_fixable:
            project.save()

    return project



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Commands related to the 'commands' section of anaconda-project.yml."""
from __future__ import absolute_import, print_function

import os
import platform
import sys

from anaconda_project.internal.cli.project_load import load_project
from anaconda_project import project_ops
from anaconda_project.internal.cli import console_utils


def _ask_command(command):
    if not console_utils.stdin_is_interactive():
        return None

    if platform.system() == 'Windows':
        other = 'windows'
    else:
        other = 'unix'
    choices = {'b': 'bokeh_app', 'c': other, 'n': 'notebook'}

    while True:
        data = console_utils.console_input("Is `{}` a (B)okeh app, (N)otebook, or (C)ommand line? ".format(command))
        data = data.lower().strip()

        if len(data) == 0 or data[0] not in choices:
            print("Please enter 'b', 'n', or 'c'.")
            print("    A Bokeh app is the project-relative path to a Bokeh script or app directory.")
            print("    A notebook file is the project-relative path to a .ipynb file.")
            print("    A command line is any command you might type at the command prompt.")
            continue

        return choices[data]


def add_command(project_dir, name, command_type, command, env_spec_name, supports_http_options):
    """Add command to anaconda-project.yml.

    Returns:
        int exit code
    """
    project = load_project(project_dir)

    command_as_filename = os.path.join(project.directory_path, command)

    if command_type is None and command.endswith(".ipynb") and os.path.isfile(command_as_filename):
        command_type = 'notebook'

    if command_type is None or command_type == 'ask':
        command_type = _ask_command(name)

    if command_type is None:  # EOF, probably not an interactive console
        print("Specify the --type option to add this command.", file=sys.stderr)
        return 1

    status = project_ops.add_command(project, name, command_type, command, env_spec_name, supports_http_options)
    if not status:
        console_utils.print_status_errors(status)
        return 1
    else:
        print("Added a command '%s' to the project. Run it with `anaconda-project run %s`." % (name, name))
        return 0


def remove_command(project_dir, name):
    """Remove a command from the project.

    Returns:
        int exit code
    """
    project = load_project(project_dir)

    status = project_ops.remove_command(project, name)
    if not status:
        console_utils.print_status_errors(status)
        return 1
    else:
        print("Removed the command '{}' from the project.".format(name))
        return 0


def list_commands(project_dir):
    """List the commands on the project.

    Returns:
        int exit code
    """
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1

    if project.commands:
        print("Commands for project: {}\n".format(project_dir))
        console_utils.print_names_and_descriptions(project.commands.values())
    else:
        print("No commands found for project: {}\n".format(project_dir))
    return 0


def main(args):
    """Submit the add command with args and returns exit code."""
    return add_command(args.directory, args.name, args.type, args.command, args.env_spec, args.supports_http_options)


def main_remove(args):
    """Submit the remove command with args and returns exit code."""
    return remove_command(args.directory, args.name)


def main_list(args):
    """Start the list command with args and return exit code."""
    return list_commands(args.directory)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Download related providers."""
from __future__ import print_function

import os
import shutil

from tornado.ioloop import IOLoop

from anaconda_project.internal.http_client import FileDownloader
from anaconda_project.internal.ziputils import unpack_zip
from anaconda_project.internal.simple_status import SimpleStatus
from anaconda_project.requirements_registry.provider import EnvVarProvider, ProviderAnalysis
from anaconda_project.provide import PROVIDE_MODE_CHECK
from anaconda_project.frontend import _new_error_recorder


class _DownloadProviderAnalysis(ProviderAnalysis):
    """Subtype of ProviderAnalysis showing if a filename exists."""

    def __init__(self, config, missing_to_configure, missing_to_provide, existing_filename):
        super(_DownloadProviderAnalysis, self).__init__(config, missing_to_configure, missing_to_provide)
        self.existing_filename = existing_filename


class DownloadProvider(EnvVarProvider):
    """Downloads a file according to the specified requirement."""

    def read_config(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
        """Override superclass to return our config."""
        config = super(DownloadProvider, self).read_config(requirement, environ, local_state_file,
                                                           default_env_spec_name, overrides)

        assert 'source' in config
        assert config['source'] != 'default'

        if config['source'] == 'unset':
            config['source'] = 'download'

        return config

    def set_config_values_as_strings(self, requirement, environ, local_state_file, default_env_spec_name, overrides,
                                     values):
        """Override superclass to clear out environ if we decide not to use it."""
        super(DownloadProvider, self).set_config_values_as_strings(requirement, environ, local_state_file,
                                                                   default_env_spec_name, overrides, values)

        if 'source' in values and values['source'] != 'environ':
            # clear out the previous setting; this is sort of a hack. The problem
            # is that we don't want to delete env vars set in actual os.environ on
            # the command line, in our first pass, and in some subtypes of EnvVarProvider
            # (CondaEnvProvider) we also don't want to use it by default. Otherwise
            # we should probably do this in EnvVarProvider. future: rethink this.
            # a possible fix is to track an initial_environ for the whole prepare
            # sequence, separately from the current running environ?
            environ.pop(requirement.env_var, None)

    def analyze(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
        """Override superclass to store additional fields in the analysis."""
        analysis = super(DownloadProvider, self).analyze(requirement, environ, local_state_file, default_env_spec_name,
                                                         overrides)
        filename = os.path.join(environ['PROJECT_DIR'], requirement.filename)
        if os.path.exists(filename):
            existing_filename = filename
        else:
            existing_filename = None
        return _DownloadProviderAnalysis(analysis.config,
                                         analysis.missing_env_vars_to_configure,
                                         analysis.missing_env_vars_to_provide,
                                         existing_filename=existing_filename)

    def _provide_download(self, requirement, context, frontend):
        filename = context.status.analysis.existing_filename
        if filename is not None:
            frontend.info("Previously downloaded file located at {}".format(filename))
            return filename

        filename = os.path.abspath(os.path.join(context.environ['PROJECT_DIR'], requirement.filename))
        if requirement.unzip:
            download_filename = filename + ".zip"
        else:
            download_filename = filename
        download = FileDownloader(url=requirement.url,
                                  filename=download_filename,
                                  hash_algorithm=requirement.hash_algorithm)

        try:
            _ioloop = IOLoop(make_current=False)
            response = _ioloop.run_sync(lambda: download.run(_ioloop))
            if response is None:
                for error in download.errors:
                    frontend.error(error)
                return None
            elif response.code == 200:
                if requirement.hash_value is not None and requirement.hash_value != download.hash:
                    frontend.error("Error downloading {}: mismatched hashes. Expected: {}, calculated: {}".format(
                        requirement.url, requirement.hash_value, download.hash))
                    return None
                if requirement.unzip:
                    unzip_errors = []
                    if unpack_zip(download_filename, filename, unzip_errors):
                        os.remove(download_filename)
                        return filename
                    else:
                        for error in unzip_errors:
                            frontend.error(error)
                        return None
                return filename
            else:
                frontend.error("Error downloading {}: response code {}".format(requirement.url, response.code))
                return None
        except Exception as e:
            frontend.error("Error downloading {}: {}".format(requirement.url, str(e)))
            return None
        finally:
            _ioloop.close()

    def provide(self, requirement, context):
        """Override superclass to start a download..

        If it locates a downloaded file with matching checksum, it sets the
        requirement's env var to that filename.

        """
        super_result = super(DownloadProvider, self).provide(requirement, context)

        if context.mode == PROVIDE_MODE_CHECK:
            return super_result
        # we do the download in both prod and dev mode

        frontend = _new_error_recorder(context.frontend)
        if requirement.env_var not in context.environ or context.status.analysis.config['source'] == 'download':
            filename = self._provide_download(requirement, context, frontend)
            if filename is not None:
                context.environ[requirement.env_var] = filename

        return super_result.copy_with_additions(errors=frontend.pop_errors())

    def unprovide(self, requirement, environ, local_state_file, overrides, requirement_status=None):
        """Override superclass to delete the downloaded file."""
        project_dir = environ['PROJECT_DIR']
        filename = os.path.abspath(os.path.join(project_dir, requirement.filename))
        try:
            if os.path.isdir(filename):
                shutil.rmtree(filename)
            elif os.path.isfile(filename):
                os.remove(filename)
            else:
                return SimpleStatus(success=True,
                                    description=("No need to remove %s which wasn't downloaded." % filename))
            return SimpleStatus(success=True, description=("Removed downloaded file %s." % filename))
        except Exception as e:
            return SimpleStatus(success=False, description=("Failed to remove %s: %s." % (filename, str(e))))
